import requests
import copy
import os
import sqlite3
import time
from pathlib import Path
from tool_config import setup_cache

script_dir = Path(__file__).parent.absolute()
database_file = script_dir / "database" / "github_commit.db"
# print("Database file: ", database_file)


conn = sqlite3.connect(database_file)
c = conn.cursor()

c.execute(
    """CREATE TABLE IF NOT EXISTS commit_data (
             api_url TEXT PRIMARY KEY,
             earliest_commit_sha TEXT,
             repo_name TEXT,
             package TEXT,
             author_login TEXT,
             author_commit_sha TEXT,
             author_login_in_1st_commit TEXT,
             author_id_in_1st_commit TEXT)"""
)

conn.commit()


# logging.info("Cache [github_cache_cache] setup complete")


github_token = os.getenv("GITHUB_API_TOKEN")
# if not github_token:
#     raise ValueError("GitHub API token is not set in the environment variables.")

headers = {
    "Authorization": f"Bearer {github_token}",
    "Accept": "application/vnd.github.v3+json",
}


def get_user_first_commit_info(data):
    """
    Get the first commit information for each author in the given data.

    Args:
        data (dict): A dictionary containing package information with authors.

    Returns:
        dict: A dictionary with updated package information including first commit details.
    """
    setup_cache("github_commits_info")

    failed_api_urls = set()

    earliest_commit_sha = None
    author_login_in_commit = None
    author_id_in_commit = None
    first_time_commit = None

    packages_data = copy.deepcopy(data)

    for package, info in packages_data.items():
        print(f"Processing {package}")
        repo_name = info["repo_name"]

        if info.get("authors"):
            for author in info.get("authors"):
                author_login = author.get("login", "No_author_login")
                commit_sha = author.get("sha", "No_commit_sha")
                author_type = author.get("a_type", "No_author_type")

                commit_result = {
                    "api_url": None,
                    "earliest_commit_sha": None,
                    "author_login_in_1st_commit": None,
                    "author_id_in_1st_commit": None,
                    "is_first_commit": None,
                    "commit_notice": None,
                }

                if author_login is None:
                    commit_result["commit_notice"] = "Author login is None"

                if "[bot]" in author_login or author_type == "Bot":
                    commit_result["earliest_commit_sha"] = "It might be a bot"
                    commit_result["commit_notice"] = "Bot author detected"

                else:
                    api_url = f"https://api.github.com/search/commits?q=repo:{repo_name}+author:{author_login}+sort:author-date-asc"

                    c.execute(
                        "SELECT earliest_commit_sha, author_login_in_1st_commit, author_id_in_1st_commit FROM commit_data WHERE api_url = ?",
                        (api_url,),
                    )
                    data = c.fetchone()

                    if data:
                        (
                            earliest_commit_sha,
                            author_login_in_commit,
                            author_id_in_commit,
                        ) = data
                        first_time_commit = True if earliest_commit_sha == commit_sha else False

                        commit_result.update(
                            {
                                "api_url": api_url,
                                "earliest_commit_sha": earliest_commit_sha,
                                "author_login_in_1st_commit": author_login_in_commit,
                                "author_id_in_1st_commit": author_id_in_commit,
                                "is_first_commit": first_time_commit,
                                "commit_notice": "Data retrieved from cache",
                            }
                        )

                    else:
                        max_retries = 2
                        base_wait_time = 2
                        retries = 0
                        success = False

                        while retries < max_retries and not success and api_url not in failed_api_urls:
                            response = requests.get(api_url, headers=headers)
                            time.sleep(2)

                            if response.status_code == 200:
                                success = True
                                commits_data = response.json()
                                earliest_commit_sha = (
                                    commits_data["items"][0]["sha"] if commits_data["items"] else None
                                )

                                author_login_in_commit = (
                                    commits_data["items"][0]["author"]["login"] if commits_data["items"] else None
                                )
                                # author_type = commits_data['items'][0]['author']['__typename'] if commits_data['items'] else None

                                author_id_in_commit = (
                                    commits_data["items"][0]["author"]["id"] if commits_data["items"] else None
                                )
                                # api_url_cache[api_url] = earliest_commit_sha

                                first_time_commit = True if earliest_commit_sha == commit_sha else False

                                c.execute(
                                    "INSERT INTO commit_data (api_url, earliest_commit_sha, repo_name, package, author_login, author_commit_sha, author_login_in_1st_commit, author_id_in_1st_commit) VALUES (?, ?, ?, ?, ?, ?, ?, ?)",
                                    (
                                        api_url,
                                        earliest_commit_sha,
                                        repo_name,
                                        package,
                                        author_login,
                                        commit_sha,
                                        author_login_in_commit,
                                        author_id_in_commit,
                                    ),
                                )
                                conn.commit()

                                commit_result.update(
                                    {
                                        "api_url": api_url,  # "https://api.github.com/search/commits?q=repo:{repo_name}+author:{author_login}+sort:author-date-asc
                                        "earliest_commit_sha": earliest_commit_sha,
                                        "author_login_in_1st_commit": author_login_in_commit,
                                        "author_id_in_1st_commit": author_id_in_commit,
                                        "is_first_commit": first_time_commit,
                                        "commit_notice": "Data retrieved from API",
                                    }
                                )

                            else:
                                print(f"Error: {response.status_code}")
                                remaining = response.headers.get("X-RateLimit-Remaining")
                                reset_time = response.headers.get("X-RateLimit-Reset")
                                wait_time = max(int(reset_time) - int(time.time()), 0)
                                print(f"Rate limit remaining: {remaining}")

                                if remaining == "0":
                                    time.sleep(wait_time)

                                else:
                                    time.sleep(base_wait_time)

                                retries += 1
                                print(f"Retrying...{retries}/{max_retries} for {api_url}")

                        if not success:
                            commit_result["commit_notice"] = f"Failed to retrieve data from API({api_url})"
                            failed_api_urls.add(api_url)

                author["commit_result"] = commit_result

        else:
            info["commit_result"] = None

    conn.close()

    return packages_data



import requests
import sqlite3
import os
import json
import time
import copy
import logging


GITHUB_TOKEN = os.getenv("GITHUB_API_TOKEN")
# if not GITHUB_TOKEN:
#     raise ValueError("GitHub API token is not set in the environment variables.")

headers = {
    "Authorization": f"Bearer {GITHUB_TOKEN}",
    "Accept": "application/vnd.github.v4+json",
}

url = "https://api.github.com/graphql"


conn = sqlite3.connect("database/github_pr_data.db")
c = conn.cursor()

c.execute(
    """CREATE TABLE IF NOT EXISTS pr_info_sample
             (package TEXT, commit_sha TEXT, commit_node_id TEXT, pr_data TEXT)"""
)

conn.commit()


def fetch_pull_requests(commit_node_id):
    query = """
    query Edges($nodeId: ID!, $first: Int) {
    node(id: $nodeId) {
        ... on Commit {
        associatedPullRequests(first: $first) {
            edges {
            node {
                author {
                login
                __typename
                }
                authorAssociation
                autoMergeRequest {
                mergeMethod
                enabledBy {
                    login
                }
                authorEmail
                }
                checksUrl
                createdAt
                mergeCommit {
                author {
                    name
                    email
                }
                }
                id
                merged
                mergedAt
                mergedBy {
                login
                __typename
                }
                number
                state
                url
                reviews(first: $first, states: APPROVED) {
                edges {
                    node {
                    author {
                        login
                        __typename
                    }
                    id
                    state
                    createdAt
                    publishedAt
                    submittedAt
                    updatedAt
                    }
                }
                }
                repository {
                name
                owner {
                    login
                    id
                }
                }
            }
            }
        }
        }
    }
    }
    """

    variables = {
        "nodeId": f"{commit_node_id}",
        "first": 5,
    }

    body = json.dumps({"query": query, "variables": variables})

    response = requests.post(url, data=body, headers=headers)

    if response.status_code != 200:
        # retry 10 sec later and try 5 times
        for i in range(5):
            print(f"Retrying in 10 seconds...")
            time.sleep(10)
            response = requests.post(url, data=body, headers=headers)
            if response.status_code == 200:
                break
        else:
            raise Exception(response.status_code, response.text)

    pr_info = response.json()

    return pr_info


def get_pr_info(data):
    logging.info("Getting PR info for commits...")

    pr_infos = []

    commits_data = copy.deepcopy(data)

    for package, info in commits_data.items():
        repo_name = info.get("repo_name")
        print(f"Checking PR info in {package}'s repository: ", repo_name)
        authors = info.get("authors", [])

        for author in authors:
            commit_sha = author.get("sha")
            commit_node_id = author.get("node_id")
            commit_url = author.get("commit_url")

            c.execute(
                "SELECT pr_data FROM pr_info_sample WHERE commit_node_id=?",
                (commit_node_id,),
            )
            result = c.fetchone()

            if result:
                pr_info = json.loads(result[0])
            else:
                if commit_node_id:
                    pr_info = fetch_pull_requests(commit_node_id)

                    c.execute(
                        "INSERT INTO pr_info_sample (package, commit_sha, commit_node_id, pr_data) VALUES (?, ?, ?, ?)",
                        (package, commit_sha, commit_node_id, json.dumps(pr_info)),
                    )
                    conn.commit()

            all_info = {
                "package": package,
                "commit_sha": commit_sha,
                "commit_node_id": commit_node_id,
                "pr_info": pr_info,
                "repo_name": repo_name,
            }
            pr_infos.append(all_info)

    return pr_infos


def get_useful_pr_info(commits_data):
    pr_infos = get_pr_info(commits_data)

    for pr_info in pr_infos:
        if pr_info:
            package = pr_info.get("package")
            commit_sha = pr_info.get("commit_sha")
            commit_node_id = pr_info.get("commit_node_id")
            repo_name = pr_info.get("repo_name")
            associated_prs = (
                pr_info.get("pr_info", {})
                .get("data", {})
                .get("node", {})
                .get("associatedPullRequests", {})
                .get("edges", [])
            )

            for author in commits_data[package].get("authors", []):
                if author.get("node_id") == commit_node_id:
                    author["commit_merged_info"] = []
                    if len(associated_prs) == 0:
                        author["commit_merged_info"].append({"merge_info": "no associated PRs"})
                    else:
                        for associated_pr in associated_prs:
                            merge_node_info = associated_pr.get("node", {})
                            author_association = merge_node_info.get("authorAssociation")
                            auto_merge_request = merge_node_info.get("autoMergeRequest", {})
                            created_at = merge_node_info.get("createdAt")
                            pr_id = merge_node_info.get("id")
                            state = merge_node_info.get("state")
                            merge_at = merge_node_info.get("mergedAt")
                            pull_url = merge_node_info.get("url")
                            reviews = merge_node_info.get("reviews", {}).get("edges", [])

                            merged_info = {
                                "repo": repo_name,
                                "commit_sha": commit_sha,
                                "commit_node_id": commit_node_id,
                                "author_association": author_association,
                                "auto_merge_request": auto_merge_request,
                                "created_at": created_at,
                                "pr_id": pr_id,
                                "state": state,
                                "merge_at": merge_at,
                                "pull_url": pull_url,
                                "reviews": [],
                            }

                            if state == "MERGED":
                                if merge_node_info.get("mergedBy"):
                                    merged_info["merge_by"] = merge_node_info.get("mergedBy", {}).get("login")
                                    merged_info["merge_by_type"] = merge_node_info.get("mergedBy", {}).get(
                                        "__typename"
                                    )
                                else:
                                    merged_info["merge_by"] = merge_node_info.get("mergedBy")

                            else:
                                merged_info["merge_by"] = None
                                merged_info["merge_by_type"] = None

                            for review in reviews:
                                review_node = review.get("node", {})
                                if review_node:
                                    if review_node.get("author", {}):
                                        review_author = review_node.get("author", {}).get("login", None)
                                        review_author_type = review_node.get("author", {}).get("__typename", None)

                                        review_info = {
                                            "review_author": review_author,
                                            "review_author_type": review_author_type,
                                            "review_state": review_node.get("state", None),
                                            "review_id": review_node.get("id", None),
                                        }

                                        merged_info["reviews"].append(review_info)

                            author["commit_merged_info"].append(merged_info)

    return commits_data



"""
This file contains the configuration for the tool.
"""

import datetime
import pathlib
import logging

import requests_cache

# change this to the install command for your project
PNPM_LIST_COMMAND = [
    "pnpm",
    "list",
    "--filter",
    "ledger-live-desktop",
    "--depth",
    "Infinity",
]


class PathManager:
    """
    Manage the paths for the results.
    """

    def __init__(self, base_dir="results"):
        self.base_dir = pathlib.Path(base_dir)

    def create_folders(self, version_tag):
        """
        Create the folders for the results.
        """

        current_time = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
        folder_name = f"results_{current_time}"
        result_folder_path = self.base_dir / folder_name
        result_folder_path.mkdir(parents=True, exist_ok=True)

        json_directory = result_folder_path / "sscs" / version_tag
        json_directory.mkdir(parents=True, exist_ok=True)
        diff_directory = result_folder_path / "diff"
        diff_directory.mkdir(parents=True, exist_ok=True)

        return result_folder_path, json_directory, diff_directory


def setup_cache(cache_name):
    """
    Setup the cache for the requests.
    """

    cache_folder = pathlib.Path("cache")
    cache_folder.mkdir(parents=True, exist_ok=True)

    cache_file = cache_folder / f"{cache_name}_cache"

    requests_cache.install_cache(
        cache_name=str(cache_file),
        backend="sqlite",
        expire_after=7776000,
        allowable_codes=(200, 301, 302, 404),
    )  # 90 days

    # logging.info(f"Cache setup complete: {cache_file}")


def setup_logger(log_file_path):
    """
    Setup the logger for the analysis.
    """

    # Set up the logger
    logger = logging.getLogger("dw_analysis")
    logger.setLevel(logging.INFO)

    # Create a console handler
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)

    # Create a file handler
    file_handler = logging.FileHandler(log_file_path)
    file_handler.setLevel(logging.INFO)

    # Create a formatter and set it for both handlers
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    console_handler.setFormatter(formatter)
    file_handler.setFormatter(formatter)

    # Add the handlers to the logger
    logger.addHandler(console_handler)
    logger.addHandler(file_handler)

    return logger



import pandas as pd
from datetime import datetime


def process_data(data):
    record = []
    record_dict = {}
    author_dict = {}

    for package_name, info in data.items():
        repo_name = info.get("repo_name", "")
        pkg_category = info.get("category", "")
        old_version = info.get("tag1", "")
        new_version = info.get("tag2", "")
        repo_link = info.get("repo_link", "")

        commits = info.get("authors", [])

        if not commits:
            record.append(
                {
                    "package_name": package_name,
                    "repo_name": repo_name,
                    "repo_link": repo_link,
                    "category": pkg_category,
                    "old_version": old_version,
                    "new_version": new_version,
                    "sha": None,
                    "author": None,
                    "author_first": None,
                    "merger": None,
                    "prr_first": None,
                    "reviewer": None,
                    "reviewer_type": None,
                }
            )

        for commit in commits:
            sha = commit.get("sha", "")
            author = commit.get("login", "")
            author_first = commit.get("commit_result", {}).get("is_first_commit", "")

            commit_merge_infos = commit.get("commit_merged_info", [])

            prr_first = None
            reviewer = None
            reviewer_type = None
            merger = None

            for commit_merge_info in commit_merge_infos:
                merger = commit_merge_info.get("merge_by", "")
                if "reviews" in commit_merge_info and isinstance(commit_merge_info["reviews"], list):
                    for review in commit_merge_info["reviews"]:
                        if isinstance(review, dict) and "prr_data" in review:
                            reviewer = review.get("review_author")
                            reviewer_type = review.get("review_author_type")
                            prr_data = review.get("prr_data", {})
                            if prr_data:
                                prr_first = prr_data.get("is_first_prr")

                if sha in record_dict:
                    if package_name not in record_dict[sha]["package_name"]:
                        record_dict[sha]["package_name"].append(package_name)
                else:
                    record_dict[sha] = {
                        "package_name": [package_name],
                        "repo_name": repo_name,
                        "old_version": old_version,
                        "new_version": new_version,
                        "sha": sha,
                        "author": author,
                        "author_first": author_first,
                        "merger": merger,
                        "prr_first": prr_first,
                        "reviewer": reviewer,
                        "reviewer_type": reviewer_type,
                    }

                package_number = len(record_dict[sha]["package_name"])
                record_dict[sha]["package_number"] = package_number

            record.append(
                {
                    "package_name": package_name,
                    "repo_name": repo_name,
                    "category": pkg_category,
                    "old_version": old_version,
                    "new_version": new_version,
                    "sha": sha,
                    "author": author,
                    "author_first": author_first,
                    "merger": merger,
                    "prr_first": prr_first,
                    "reviewer": reviewer,
                    "reviewer_type": reviewer_type,
                }
            )

    record_list = list(record_dict.values())
    author_list = list(author_dict.values())

    return record, record_list, author_list


def create_dataframe(record):
    df = pd.DataFrame(record)
    return df


def filter_df(df):
    df_author_first = df[(df["author_first"] == True)]
    df_review_first = df[df["prr_first"] == True]
    df_both_first = df[(df["author_first"] == True) & (df["prr_first"] == True)]

    return df_author_first, df_review_first, df_both_first


def generate_diff_report(data, project_repo_name, release_version_old, release_version_new, output_file):
    print(f"Generating differential report for {project_repo_name}")
    record, record_list, author_list = process_data(data)

    df_all = create_dataframe(record)

    # To know downgraded packages
    downgraded = df_all["category"] == "Downgraded package"
    downgraded_number = df_all[downgraded]["package_name"].nunique()
    selected_columns = [
        "package_name",
        "repo_link",
        "category",
        "old_version",
        "new_version",
    ]
    df_down_selected = df_all[downgraded][selected_columns]
    df_down_selected_without_index = df_down_selected.reset_index(drop=True)

    # We want a table that indexes the author
    df_author = pd.DataFrame(record_list).set_index("author")

    new_order = ["sha"] + [col for col in df_author.columns if col != "sha"]
    df_author = df_author.reindex(columns=new_order)

    df_author_new_author, df_author_new_reviewer, df_author_both_new = filter_df(df_author)

    # know the columns

    profile_url = "https://github.com/"

    cp_df_author_new_author = df_author_new_author.copy()
    cp_df_author_new_author.index = df_author_new_author.index.map(lambda x: f"[{x}]({profile_url}{x})")
    # df_author_new_author.loc[:,"author"] = df_author_new_author.loc[:,"author"].map(lambda x: f"[{x}]({profile_url}{x})")

    cp_df_author_new_reviewer = df_author_new_reviewer.copy()
    cp_df_author_new_reviewer["reviewer"] = df_author_new_reviewer["reviewer"].apply(
        lambda x: f"[{x}]({profile_url}{x})"
    )

    cp_df_author_both_new = df_author_both_new.copy()
    cp_df_author_both_new.index = df_author_both_new.index.map(lambda x: f"[{x}]({profile_url}{x})")
    cp_df_author_both_new["reviewer"] = df_author_both_new["reviewer"].apply(lambda x: f"[{x}]({profile_url}{x})")

    new_author_commits = df_author_new_author.shape[0]
    new_reviewer_commits = df_author_new_reviewer.shape[0]
    both_new_commits = df_author_both_new.shape[0]

    counts = {
        ":heavy_exclamation_mark: Downgraded packages": downgraded_number,
        ":alien: Commits made by both New Authors and Reviewers": both_new_commits,
        ":neutral_face: Commits made by New Authors": new_author_commits,
        ":see_no_evil: Commits approved by New Reviewers": new_reviewer_commits,
    }

    # We write into a markdown file
    with open(output_file, "w", encoding="utf-8") as f:
        f.write(f"# Differential Report of {project_repo_name} - {release_version_old} & {release_version_new}\n")
        f.write("\n")

        for key, val in counts.items():
            f.write(f"\n {key}: {val}\n")
            f.write("\n")

        f.write("### Fine grained information\n")

        if downgraded_number > 0:
            f.write("\n")
            f.write(
                f"""
<details>
            <summary>Downgraded packages</summary>
                        """
            )
            f.write("\n\n\n")
            f.write(df_down_selected_without_index.to_markdown(index=True))
            f.write("\n")
            f.write("</details>")
            f.write("\n")

        if both_new_commits > 0:
            f.write("\n")
            f.write(
                f"""
<details>
            <summary>Both Authors and Reviewers are new to the repository </summary>
                """
            )
            f.write("\n\n\n")
            f.write(cp_df_author_both_new.to_markdown(index=True))
            f.write("\n")
            f.write("</details>")

        if new_author_commits > 0:
            f.write("\n")
            f.write(
                f"""
<details>
            <summary>Authors are new to the repository </summary>
                    """
            )
            f.write("\n\n\n")
            f.write(cp_df_author_new_author.to_markdown(index=True))
            f.write("\n")
            f.write("</details>")

        if new_reviewer_commits > 0:
            f.write("\n")
            f.write(
                f"""
<details>
            <summary>Reviewers are new to the repository </summary>
                """
            )
            f.write("\n\n\n")
            f.write(cp_df_author_new_reviewer.to_markdown(index=True))
            f.write("\n")
            f.write("</details>")

        f.write(f"---\n")
        f.write(
            f"\nReport created by [dirty-waters](https://github.com/chains-project/dirty-waters/) - version: commit.\n"
        )
        f.write(f"\nReport created on {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}\n")

        # Tool version
        # tool_commit_hash = subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD']).strip().decode('utf-8')
        # md_file.write(f"- Tool version: {tool_commit_hash}\n")
        f.write(f"- project Name: {project_repo_name}\n")
        f.write(f"- Compared project Versions: {release_version_old} & {release_version_new}\n")
    print(f"Report generated at {output_file}")



"""
Generate the report for the static analysis results.
"""

import json
import subprocess
from datetime import datetime
import pandas as pd

# Mapping smell to package managers that support it
SUPPORTED_SMELLS = {
    "no_source_code": ["yarn-classic", "yarn-berry", "pnpm", "npm", "maven"],
    "github_404": ["yarn-classic", "yarn-berry", "pnpm", "npm", "maven"],
    "release_tag_not_found": ["yarn-classic", "yarn-berry", "pnpm", "npm", "maven"],
    "deprecated": ["yarn-classic", "yarn-berry", "pnpm", "npm"],
    "forked_package": ["yarn-classic", "yarn-berry", "pnpm", "npm", "maven"],
    "provenance": ["yarn-classic", "yarn-berry", "pnpm", "npm"],
}


def load_data(filename):
    """Load data from a JSON file got from static analysis."""

    with open(filename, encoding="utf-8") as f:
        return json.load(f)


def create_dataframe(data):
    """
    Create a dataframe from the data got from static analysis.

    """

    rows = []

    for package_name, package_data in data.items():
        github_exists_data = package_data.get("github_exists", {}) or {}

        match_data = package_data.get("match_info", {}) or {}
        release_tag_exists_info = github_exists_data.get("release_tag", {}) or {}

        # Create a row for each package

        row = {
            "package_name": package_name,
            "deprecated_in_version": package_data.get("package_info", {}).get("deprecated_in_version"),
            "provenance_in_version": package_data.get("package_info", {}).get("provenance_in_version"),
            "all_deprecated": package_data.get("package_info", {}).get("all_deprecated", None),
            "github_url": github_exists_data.get("github_url", "Could not find repo from package registry"),
            "github_exists": github_exists_data.get("github_exists", None),
            "github_redirected": github_exists_data.get("github_redirected", None),
            "archived": github_exists_data.get("archived", None),
            "is_fork": github_exists_data.get("is_fork", None),
            "parent_repo_link": github_exists_data.get("parent_repo_link", None),
            "forked_from": github_exists_data.get("parent_repo_link", "-"),
            "open_issues_count": github_exists_data.get("open_issues_count", "-"),
            "is_match": match_data.get("match", None),
            # "release_tag_exists_info": github_exists_data.get("release_tag", {}),
            "release_tag_exists": release_tag_exists_info.get("exists", "-"),
            "tag_version": release_tag_exists_info.get("tag_version", "-"),
            "tag_url": release_tag_exists_info.get("url", "-"),
            "tag_related_info": release_tag_exists_info.get("tag_related_info", "-"),
            "status_code_for_release_tag": release_tag_exists_info.get("status_code", "-"),
        }
        rows.append(row)

    df = pd.DataFrame(rows)

    return df.set_index("package_name")


def write_summary(df, project_name, release_version, package_manager, filename, mode="w"):
    """
    Write a summary of the static analysis results to a markdown file.
    """

    no_source_code_repo_df = df.loc[df["github_url"] == "No_repo_info_found", ["github_url", "github_exists"]]
    github_repo_404_df = df.loc[df["github_exists"] == False, ["github_url", "github_exists"]]

    combined_repo_problems_df = (
        pd.concat([no_source_code_repo_df, github_repo_404_df])
        .reset_index(drop=False)
        .drop_duplicates(subset=["package_name"])
    )
    # could not find release tag while github exists
    release_tag_not_found_df = df.loc[
        (df["release_tag_exists"] == False) & (df["github_exists"] == True),
        [
            "release_tag_exists",
            "tag_version",
            "github_url",
            "tag_related_info",
            "status_code_for_release_tag",
        ],
    ]
    # all_deprecated_df = df[df["all_deprecated"] is True]
    version_deprecated_df = df.loc[
        df["deprecated_in_version"] == True,
        [
            "deprecated_in_version",
            "all_deprecated",
        ],
    ]
    forked_package_df = df.loc[
        df["is_fork"] == True,
        [
            "is_fork",
            "parent_repo_link",
        ],
    ]
    provenance_df = df.loc[
        df["provenance_in_version"] == False,
        [
            "provenance_in_version",
        ],
    ]

    common_counts = {
        "### Total packages in the supply chain": len(df),
    }

    warning_counts = {
        ":heavy_exclamation_mark: Packages with no Source Code URL(⚠️⚠️⚠️)": (
            df["github_url"] == "No_repo_info_found"
        ).sum(),
        ":no_entry: Packages with Github URLs that are 404(⚠️⚠️⚠️)": (df["github_exists"] == False).sum(),
        ":wrench: Packages with accessible source code repos but inaccessible GitHub tags(⚠️⚠️⚠️)": (
            release_tag_not_found_df.shape[0]
        ),
        ":x: Packages that are deprecated(⚠️⚠️)": (df["deprecated_in_version"] == True).sum(),
        ":cactus: Packages that are forks(⚠️⚠️)": (df["is_fork"] == True).sum(),
        ":black_square_button: Packages without provenance(⚠️)": (df["provenance_in_version"] == False).sum(),
    }

    not_on_github_counts = (df["github_url"] == "Not_github_repo").sum()

    source_sus = (df["github_url"] == "No_repo_info_found").sum() + (df["github_exists"] == False).sum()

    with open(filename, mode, encoding="utf-8") as md_file:
        md_file.write(f"# Software Supply Chain Report of {project_name} - {release_version}\n")
        md_file.write("\n")

        md_file.write(
            """
<details>
    <summary>How to read the results :book: </summary>
    \n Dirty-waters has analyzed your project dependencies and found different categories for each of them:\n
    \n - ⚠️⚠️⚠️ : high severity \n
    \n - ⚠️⚠️: medium severity \n
    \n - ⚠️: low severity \n
</details>
        """
        )

        md_file.write("\n")

        for key, val in common_counts.items():
            md_file.write(f"\n {key}: {val}\n")
        md_file.write("\n")

        for key, val in warning_counts.items():
            md_file.write(f"\n{key}: {val}\n")
        md_file.write("\n")

        # md_file.write(f"#### Other info")

        if package_manager not in SUPPORTED_SMELLS["no_source_code"]:
            md_file.write(
                f"""
<details>
    <summary>Other info:</summary>
    \n- Source code repo is not hosted on github:  {not_on_github_counts} \n
</details>
                        
                        """
            )
            md_file.write("\n\n")

        md_file.write("\n### Fine grained information\n")
        md_file.write(
            "\n:dolphin: For further information about software supply chain smells in your project, take a look at the following tables.\n"
        )

        if not combined_repo_problems_df.empty:
            md_file.write(
                f"""
<details>
    <summary>Source code links that could not be found({source_sus})</summary>
        """
            )
            md_file.write("\n\n\n")
            combined_repo_problems_df.index = range(1, len(combined_repo_problems_df) + 1)
            markdown_text = combined_repo_problems_df.reset_index().to_markdown(index=False)
            md_file.write(markdown_text)
            md_file.write("\n</details>")
        elif package_manager not in SUPPORTED_SMELLS["github_404"]:
            md_file.write(
                f"\nThe package manager ({package_manager}) does not support checking for not found source code links.\n"
            )
        else:
            md_file.write("All analyzed packages have a source code repo.\n")

        if not release_tag_not_found_df.empty:
            md_file.write(
                f"""

<details>
    <summary>List of packages with available source code repos but with inaccessible tags({(release_tag_not_found_df.shape[0])})</summary>
        """
            )
            md_file.write("\n\n\n")
            markdown_text = release_tag_not_found_df.reset_index().to_markdown(index=False)
            md_file.write(markdown_text)
            md_file.write("\n</details>")
        elif package_manager not in SUPPORTED_SMELLS["release_tag_not_found"]:
            md_file.write(
                f"\nThe package manager ({package_manager}) does not support checking for inaccessible tags.\n"
            )
        else:
            md_file.write("\nAll packages have accessible tags.\n")

        if not version_deprecated_df.empty:
            md_file.write(
                f"""
<details>
    <summary>List of deprecated packages({(df['deprecated_in_version'] == True).sum()})</summary>
        """
            )
            md_file.write("\n\n\n")
            markdown_text = version_deprecated_df.reset_index().to_markdown(index=False)
            md_file.write(markdown_text)
            md_file.write("\n</details>")
        elif package_manager not in SUPPORTED_SMELLS["deprecated"]:
            md_file.write(
                f"\nThe package manager ({package_manager}) does not support checking for deprecated packages.\n"
            )
        else:
            md_file.write("\nNo deprecated package found.\n")

        if not forked_package_df.empty:
            md_file.write(
                f"""

<details>
    <summary>List of packages from fork({(df["is_fork"] == True).sum()}) </summary>
        """
            )
            md_file.write("\n\n\n")
            markdown_text = forked_package_df.reset_index().to_markdown(index=False)
            md_file.write(markdown_text)
            md_file.write("\n</details>\n")
        elif package_manager not in SUPPORTED_SMELLS["forked_package"]:
            md_file.write(
                f"\nThe package manager ({package_manager}) does not support checking for forked packages.\n"
            )
        else:
            md_file.write("\nNo package is from fork.\n")

        # TODO: they're printing irrelevant information in each table, I think
        if not provenance_df.empty:
            md_file.write(
                f"""
<details>
    <summary>List of packages without provenance({(df["provenance_in_version"] == False).sum()})</summary>
        """
            )
            md_file.write("\n\n\n")
            markdown_text = provenance_df.reset_index().to_markdown(index=False)
            md_file.write(markdown_text)
            md_file.write("\n</details>\n")
        elif package_manager not in SUPPORTED_SMELLS["provenance"]:
            md_file.write(f"\nThe package manager ({package_manager}) does not support checking for provenance.\n")
        else:
            md_file.write("\nAll packages have provenance.\n")

        md_file.write("\n### Call to Action:\n")
        md_file.write(
            """
                      
<details>
    <summary>👻What do I do now? </summary>
        For packages without source code & accessible release tags:  \n
        Pull Request to the maintainer of dependency, requesting correct repository metadata and proper tagging. \n
        \nFor deprecated packages:\n
        1. Confirm the maintainer’s deprecation intention 
        2. Check for not deprecated versions
        \nFor packages without provenance:\n
        Open an issue in the dependency’s repository to request the inclusion of provenance and build attestation in the CI/CD pipeline. 
        \nFor packages that are forks\n
        Inspect the package and its GitHub repository to verify the fork is not malicious.
</details>



"""
        )
        md_file.write("---\n")
        md_file.write("\nReport created by [dirty-waters](https://github.com/chains-project/dirty-waters/).\n")
        md_file.write(f"\nReport created on {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}\n")

        # Tool version
        tool_commit_hash = subprocess.check_output(["git", "rev-parse", "--short", "HEAD"]).strip().decode("utf-8")
        md_file.write(f"- Tool version: {tool_commit_hash}\n")
        md_file.write(f"- Project Name: {project_name}\n")
        md_file.write(f"- Project Version: {release_version}\n")


def get_s_summary(data, project_name, release_version, package_manager, summary_filename):
    """
    Get a summary of the static analysis results.
    """

    df = create_dataframe(data)
    write_summary(df, project_name, release_version, package_manager, filename=summary_filename, mode="w")
    print(f"Report created at {summary_filename}")



import json
import os
import base64
import time
import urllib.parse
from tqdm import tqdm

import requests

import tool_config
from compare_commits import tag_format as construct_tag_format


github_token = os.getenv("GITHUB_API_TOKEN")

headers = {
    "Authorization": f"Bearer {github_token}",
    "Accept": "application/vnd.github.v3+json",
}

# tool_config.setup_cache("static")


MAX_WAIT_TIME = 15 * 60


def check_deprecated_and_provenance(package, package_version, pm):
    """
    Check if the package is deprecated and if it has a provenance from the package manager's registry.
    """

    def check_npm(package, package_version):
        try:
            response = requests.get(f"https://registry.npmjs.org/{package}", timeout=20)

            response.raise_for_status()
        except requests.RequestException:
            return {
                "package_only_name": package,
                "package_version": package_version,
                "error": "Failed to fetch package data",
                "status_code": response.status_code if response else "No Response",
            }

        data = response.json()

        if_deprecated = False
        has_provenance = False
        provenance_url = None
        provenance_info = None
        all_deprecated = True

        version_info = data.get("versions", {}).get(package_version, {})
        all_versions = data.get("versions", {})

        deprecated_in_version = version_info.get("deprecated", "")
        provenance_in_version = version_info.get("dist", {}).get("attestations", "")

        if deprecated_in_version:
            if_deprecated = True

        if provenance_in_version:
            has_provenance = True
            provenance_url = provenance_in_version.get("url")
            provenance_info = provenance_in_version.get("provenance")

        for version in all_versions.values():
            if not version.get("deprecated"):
                all_deprecated = False
                break

        npm_package_info = {
            "package_only_name": package,
            "package_version": package_version,
            "deprecated_in_version": if_deprecated,
            "provenance_in_version": has_provenance,
            "all_deprecated": all_deprecated,
            "provenance_url": provenance_url,
            "provenance_info": provenance_info,
            "status_code": 200,
        }

        return npm_package_info

    def check_maven(package, package_version):
        return {
            "package_only_name": package,
            "package_version": package_version,
            "status_code": 404,
            "error": "Maven does not have a registry",
        }

    if pm in ("yarn-berry", "yarn-classic", "pnpm", "npm"):
        return check_npm(package, package_version)
    elif pm == "maven":
        # maven doesn't have this
        return check_maven(package, package_version)
    else:
        # log stuff
        # blow up
        logging.error(f"Package manager {pm} not supported.")


def api_constructor(package_name, repository):
    repo_url = repository.replace("https://", "").replace("http://", "").replace("/issues", "")

    # simplified_path = repo_url.replace("github.com/", "").split('#')[0].split('tree/master')[0].rstrip('/')
    simplified_path = (
        repo_url.replace("github.com:", "github.com/")
        .replace("github.com/", "")
        .split("#")[0]
        .split("/tree")[0]
        .rstrip("/")
    )

    if simplified_path.endswith(".git"):
        simplified_path = simplified_path[:-4]

    repo_api = f"https://api.github.com/repos/{simplified_path}"

    error_message = None

    try:
        parts = package_name.split("@")
        package_full_name = None
        name = None
        version = None

        if len(parts) > 2 or package_name.startswith("@"):
            package_full_name = f"@{parts[1]}"
            _, name = parts[1].split("/")  # scope,name
            version = parts[2]

        elif len(parts) == 2:
            if "/" in parts[1]:
                package_full_name = parts[0]
                scope, name = parts[0].split("/")
                version = parts[1]
            else:
                package_full_name = parts[0]
                name = parts[0]
                version = parts[1]

    except (ValueError, TypeError, AttributeError) as e:
        error_message = f"Error: {str(e)}"
    except Exception as e:
        error_message = f"Unexpected error: {str(e)}"

    return repo_api, simplified_path, package_full_name, name, version, error_message


def make_github_request(url, headers):
    """Make a GET request to the GitHub API."""

    response = requests.get(url, headers=headers)

    if response.status_code == 403 and int(response.headers.get("X-RateLimit-Remaining", 0)) <= 10:
        reset_time = int(response.headers.get("X-RateLimit-Reset", 0))
        sleep_time = min(reset_time - int(time.time()), MAX_WAIT_TIME)
        print(f"\nRate limit reached. Waiting for {sleep_time} seconds.")
        time.sleep(sleep_time)
        print("\nResuming analysis...")
        response = requests.get(url, headers=headers)
    return response


def check_existence(package_name, repository):
    """Check if the package exists in the repository."""
    repo_api, simplified_path, package_full_name, _, version, error_message = api_constructor(package_name, repository)

    repo_link = f"https://github.com/{simplified_path}".lower()
    github_exists = False
    archived = False
    is_fork = False
    release_tag_exists = False
    release_tag_url = None
    tag_related_info = None
    status_code_release_tag = None
    github_redirected = False
    now_repo_url = None
    open_issues_count = None

    response = make_github_request(repo_api, headers=headers)
    status_code = response.status_code
    data = response.json()

    if status_code != 200:
        print(f"Error: {data.get('message', 'No message')}")
        archived = None
        is_fork = None
        repo_link = f"https://github.com/{simplified_path}".lower()

    if status_code == 200:
        github_exists = True
        open_issues_count = data["open_issues"]
        if data["archived"]:
            archived = True
        if data["fork"]:
            is_fork = True
            # Get the parent repo link
            parent_repo_link = data.get("parent", {}).get("html_url", "")

        now_repo_url = data.get("html_url", "").lower()
        repo_link = f"https://github.com/{simplified_path}".lower()

        if now_repo_url != repo_link:
            github_redirected = True
        else:
            now_repo_url = None

        # check if the repo has any tags(to reduce the number of API requests)
        have_no_tags_check_api = f"{repo_api}/tags"
        have_no_tags_response = requests.get(have_no_tags_check_api, headers=headers)
        have_no_tags_response_status_code = have_no_tags_response.status_code
        have_no_tags_data = have_no_tags_response.json()

        if len(have_no_tags_data) == 0:
            release_tag_exists = False
            release_tag_url = None
            tag_related_info = "No tag found in the repo"
            status_code_release_tag = have_no_tags_response_status_code

        else:
            tag_possible_formats = construct_tag_format(version, package_full_name)

            if tag_possible_formats:
                for tag_format in tag_possible_formats:
                    tag_url = f"{repo_api}/git/ref/tags/{tag_format}"
                    response = make_github_request(tag_url, headers=headers)
                    if response.status_code == 200:
                        release_tag_exists = True
                        release_tag_url = tag_url
                        tag_related_info = f"Tag {tag_format} is found in the repo"
                        status_code_release_tag = response.status_code
                        break
                    else:
                        tag_related_info = "Tags are not found in the repo"
                        status_code_release_tag = response.status_code

    github_info = {
        "github_api": repo_api,
        "github_url": repo_link,
        "github_exists": github_exists,
        "github_redirected": github_redirected,
        "redirected_repo": now_repo_url,
        "status_code": status_code,
        "archived": archived,
        "is_fork": is_fork,
        "release_tag": {
            "exists": release_tag_exists,
            "tag_version": version,
            "url": release_tag_url,
            "tag_related_info": tag_related_info,
            "status_code": status_code_release_tag,
        },
        "parent_repo_link": parent_repo_link if is_fork else None,
        "open_issues_count": open_issues_count,
        "error": error_message if error_message else "No error message.",
    }

    return github_info


def get_api_content(api, headers):
    """
    Get the content of the API.
    """
    try:
        response = requests.get(api, headers=headers)
        response.raise_for_status()
        content = response.json()
        file_content = base64.b64decode(content["content"]).decode("utf-8")
        package_json = json.loads(file_content)
        return package_json

    except (
        requests.HTTPError,
        requests.ConnectionError,
        requests.Timeout,
        json.JSONDecodeError,
    ) as e:
        print(f"Request error: {str(e)} for URL: {api}")
        return None
    except Exception as e:
        print(f"Unexpected error: {str(e)} for URL: {api}")
        return None


def check_name_match_for_fork(package_name, repository):
    name_match = False
    unmatch_info = None

    repo_api, repo, _, pkg_only_name, _, _ = api_constructor(package_name, repository)
    fork_api_json_url = [
        f"{repo_api}/contents/package.json",
        f"{repo_api}/contents/contents/packages/{pkg_only_name}/package.json",
    ]

    fork_api_json = None
    api = ""
    status_code = None

    for api in fork_api_json_url:
        fork_api_response = requests.get(api, headers=headers, timeout=20)
        status_code = fork_api_response.status_code
        if status_code == 200:
            fork_api_json = fork_api_response.json()
            break

    if not fork_api_json:
        return {
            "repository": repository,
            "find_json_package": "",
            "repo_name": repo,
            "package_api_in_packages": api,
            "match": False,
            "unmatch_info": {
                "status_code": status_code,
                "reason": "API requests failed",
                "api_url": api,
            },
        }

    file_content = base64.b64decode(fork_api_json["content"]).decode("utf-8")
    package_json_file = json.loads(file_content)

    package_name_in_fork = package_json_file.get("name", "")
    if package_name_in_fork == package_name.rsplit("@", 1)[0]:
        name_match = True

    if not name_match:
        unmatch_info = {
            "status_code": status_code,
            "repo_name": repository,
            # 'total_count': search_results['total_count'],
            "package_name": package_name,
            "find_json_package": package_name_in_fork,
            "reason": "could not find",
            "api_url": api,
        }

    match_info = {
        "repository": repository,
        "find_json_package": package_name_in_fork,
        "package_name": package_name.rsplit("@", 1)[0],
        "repo_name": repo,
        "package_api_in_packages": api,
        "match": name_match,
        "unmatch_info": None if name_match else unmatch_info,
    }

    return match_info


def check_name_match(package_name, repository):
    tool_config.setup_cache("check_name")
    # logging.info("Cache [check_name_cache] setup complete")

    _, repo_name, _, _, _, _ = api_constructor(package_name, repository)
    original_package_name = package_name.rsplit("@", 1)[0]

    query = f'"\\"name\\": \\"{original_package_name}\\""'
    repo = f"{repo_name}"

    encoded_search_term = urllib.parse.quote(query)
    encoded_repo = urllib.parse.quote(repo)

    url = f"https://api.github.com/search/code?q={encoded_search_term}+in:file+repo:{encoded_repo}"

    response = requests.get(url, headers=headers, timeout=20)

    if not response.from_cache:
        time.sleep(6)

    status_code = response.status_code

    is_match = False
    unmatch_info = None
    package_api_in_packages = None

    if status_code != 200:
        search_results = None

    if status_code == 200:
        search_results = response.json()
        if search_results["total_count"] == 0:
            package_name = "can not find"

        else:
            for item in search_results["items"]:
                if item["name"] == "package.json":
                    package_api_in_packages = item["url"]
                    is_match = True

    else:
        is_match = False

    if not is_match:
        unmatch_info = {
            "status_code": status_code,
            "repo_name": repository,
            # 'total_count': search_results['total_count'],
            "package_name": package_name,
            "search_term": original_package_name,
            "query": query,
            "reason": "could not find",
            "api_url": url,
        }

    match_info = {
        "repository": repository,
        "find_json_package": package_api_in_packages,
        "repo_name": repo,
        "package_api_in_packages": package_api_in_packages,
        "match": is_match,
        "unmatch_info": None if is_match else unmatch_info,
    }

    return match_info


def analyze_package_data(package, repo_url, pm, check_match=False):
    package_info = {
        "deprecated": None,
        "provenance": None,
        "package_info": None,
        "github_exists": None,
        "match_info": None,
    }

    try:
        # TODO: check if this needs to be different because of differences between npm, maven, etc
        package_name, package_version = package.rsplit("@", 1)
        package_infos = check_deprecated_and_provenance(package_name, package_version, pm)
        package_info["deprecated"] = package_infos.get("deprecated_in_version")
        package_info["provenance"] = package_infos.get("provenance_in_version")
        package_info["package_info"] = package_infos

        if "Could not find" in repo_url:
            package_info["github_exists"] = {"github_url": "No_repo_info_found"}
        elif "not github" in repo_url:
            package_info["github_exists"] = {"github_url": "Not_github_repo"}
        else:
            github_info = check_existence(package, repo_url)
            package_info["github_exists"] = github_info
            if github_info.get("github_exists"):
                repo_url_to_use = github_info.get("redirected_repo") or repo_url
                if check_match:
                    # TODO: why do we only check this is provenance is false??
                    # TODO: maven currently not supported for this because of the above
                    if package_info["provenance"] == False:
                        if github_info.get("is_fork") == True or github_info.get("archived") == True:
                            package_info["match_info"] = check_name_match_for_fork(package, repo_url_to_use)
                        else:
                            package_info["match_info"] = check_name_match(package, repo_url_to_use)
                    else:
                        package_info["match_info"] = {
                            "has_provenance": True,
                            "match": True,
                            "repo_name": repo_url_to_use,
                        }

    except (ValueError, TypeError, AttributeError) as e:
        return None, {
            "error_type": type(e).__name__,
            "message": str(e),
            "repo_url": repo_url,
            "package": package,
            "package_info": package_info,
        }

    return package_info, None


def get_static_data(folder, packages_data, pm, check_match=False):
    print("Analyzing package static data...")
    package_all = {}
    errors = {}

    with tqdm(total=len(packages_data), desc="Analyzing packages") as pbar:
        for package, repo_urls in packages_data.items():
            # print(f"Analyzing {package}")
            tqdm.write(f"{package}")
            repo_url = repo_urls.get("github", "")
            analyzed_data, error = analyze_package_data(package, repo_url, pm, check_match=check_match)
            pbar.update(1)

            if error:
                errors[package] = error
            else:
                package_all[package] = analyzed_data

    # filepaths

    # file_path = os.path.join(folder, "all_info.json")
    # error_path = os.path.join(folder, "errors.json")

    # save_results_to_file(file_path, package_all)
    # save_results_to_file(error_path, errors)

    return package_all, errors


def save_results_to_file(filename, data):
    with open(filename, "w", encoding="utf-8") as file:
        json.dump(data, file, indent=4)



import os
import subprocess
import re
import json
import sqlite3
import logging
from pathlib import Path
from tqdm import tqdm

# from datetime import datetime


TIMEOUT = 60

script_dir = Path(__file__).parent.absolute()
database_file = script_dir / "database" / "github_repo_info_all.db"

conn = sqlite3.connect(database_file)
c = conn.cursor()

c.execute(
    """CREATE TABLE IF NOT EXISTS pkg_github_repo_output (
                package TEXT PRIMARY KEY,
                github TEXT)"""
)

conn.commit()


def write_output(folder_path, filename, data):
    folder_path = os.path.join(folder_path, "dep_lists")
    os.makedirs(folder_path, exist_ok=True)

    file_path = os.path.join(folder_path, filename)
    # check if the folder exists

    with open(file_path, "w", encoding="utf-8") as f:
        if isinstance(data, list):
            f.write("\n".join(data))
            f.write(f"\nTotal: {len(data)}\n")
        else:
            json.dump(data, f, indent=2)


def extract_repo_url(repo_info):
    pattern = r"(github.*)"
    match = re.search(pattern, repo_info, re.IGNORECASE)
    return match.group(1) if match else "not github"


def process_package(
    package,
    pm,
    repos_output,
    undefined,
    same_repos_deps,
    some_errors,
    repos_output_json,
):
    c.execute("SELECT github FROM pkg_github_repo_output WHERE package = ?", (package,))
    db_result = c.fetchone()

    if db_result:
        repo_info = db_result[0]

    else:
        try:
            if pm == "yarn-berry" or pm == "yarn-classic":
                command = ["yarn", "info", package, "repository.url"]
                result = subprocess.run(
                    command,
                    capture_output=True,
                    text=True,
                    check=True,
                    timeout=TIMEOUT,
                )

            elif pm == "pnpm":
                command = ["pnpm", "info", package, "repository.url"]
                result = subprocess.run(
                    command,
                    capture_output=True,
                    text=True,
                    check=True,
                    timeout=TIMEOUT,
                )

            elif pm == "npm":
                command = ["npm", "info", package, "repository.url"]
                result = subprocess.run(
                    command,
                    capture_output=True,
                    text=True,
                    check=True,
                    timeout=TIMEOUT,
                )

            elif pm == "maven":
                # package is in the form of group_id:artifact_id@version -- we need all 3
                name, version = package.split("@")
                group_id, artifact_id = name.split(":")
                command = [
                    "mvn",
                    "help:evaluate",
                    "-Dexpression=project.scm.url",
                    f"-Dartifact={group_id}:{artifact_id}:{version}",
                    "-q",
                    "-DforceStdout",
                ]
                result = subprocess.run(
                    command,
                    capture_output=True,
                    text=True,
                    check=True,
                    timeout=TIMEOUT,
                )

            else:
                raise ValueError(f"Unsupported package manager: {pm}")

            repo_info = result.stdout if result.stdout else result.stderr
            # print(f"Repo info for {package}: {repo_info}")
            c.execute(
                "INSERT OR IGNORE INTO pkg_github_repo_output (package, github) VALUES (?,?)",
                (package, repo_info),
            )
            conn.commit()

        except subprocess.TimeoutExpired:
            logging.error(
                f"Command {command} timed out after {TIMEOUT} seconds for package {package}",
            )
            repo_info = None

        except subprocess.CalledProcessError as e:
            logging.error(f"Command {command} failed for package {package}: {e}")
            repo_info = None

    # TODO: npm?
    package = package.replace("@npm:", "@")

    if (
        repo_info is None
        or "Undefined" in repo_info
        or "undefined" in repo_info
        or "ERR!" in repo_info
        # or "error" in repo_info
    ):
        repos_output_json[package] = {"github": "Could not find"}
        undefined.append(f"Undefined for {package}, {repo_info}")
    else:
        url = extract_repo_url(repo_info)
        # print(f"[INFO] Found GitHub URL for {package}: {url}")
        repos_output_json[package] = {"github": url}
        if url:
            repos_output.append(url)
            if url not in same_repos_deps:
                same_repos_deps[url] = []
            same_repos_deps[url].append(package)
        else:
            some_errors.append(f"No GitHub URL for {package}\n{repo_info}")


def get_github_repo_url(folder, dep_list, pm):
    repos_output = []  # List to store GitHub URLs
    some_errors = []  # List to store packages without a GitHub URL
    undefined = []  # List to store packages with undefined repository URLs
    same_repos_deps = {}  # Dict to store packages with same GitHub URL
    repos_output_json = {}  # Dict to store packages with GitHub URL

    print("Getting GitHub URLs of packages...")
    total_packages_to_process = len(dep_list.get("resolutions", []))
    # have not process patches
    with tqdm(total=total_packages_to_process, desc="Getting GitHub URLs") as pbar:
        for pkg_res in dep_list.get("resolutions"):
            package = pkg_res

            process_package(
                package,
                pm,
                repos_output,
                undefined,
                same_repos_deps,
                some_errors,
                repos_output_json,
            )
            pbar.update(1)

    # Write collected data to files
    unique_repos_output = sorted(set(repos_output))
    outputs = {
        # "github_repo_all.txt": repos_output,
        # "github_repo_unique.txt": unique_repos_output,
        "github_repo_undefined.log": undefined,
        # "github_repo_some_error.log": some_errors,
        "github_repos_depsnsamerepo.json": same_repos_deps,
        "github_repos_output.json": repos_output_json,
    }

    for filename, data in outputs.items():
        write_output(folder, filename, data)

    return repos_output_json
